# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The RPC-invocation-side bridge between RPC Framework and GRPC-on-the-wire."""

import enum
import logging
import threading
import time

from grpc._adapter import _common
from grpc._adapter import _intermediary_low as _low
from grpc.framework.base import interfaces as base_interfaces
from grpc.framework.base import null
from grpc.framework.foundation import activated
from grpc.framework.foundation import logging_pool

_THREAD_POOL_SIZE = 10

_INVOCATION_EVENT_KINDS = (
    _low.Event.Kind.METADATA_ACCEPTED,
    _low.Event.Kind.FINISH
)


@enum.unique
class _LowWrite(enum.Enum):
  """The possible categories of low-level write state."""

  OPEN = 'OPEN'
  ACTIVE = 'ACTIVE'
  CLOSED = 'CLOSED'


class _RPCState(object):
  """The full state of any tracked RPC.

  Attributes:
    call: The _low.Call object for the RPC.
    outstanding: The set of Event.Kind values describing expected future events
      for the RPC.
    active: A boolean indicating whether or not the RPC is active.
    common: An _common.RPCState describing additional state for the RPC.
  """

  def __init__(self, call, outstanding, active, common):
    self.call = call
    self.outstanding = outstanding
    self.active = active
    self.common = common


def _write(operation_id, call, outstanding, write_state, serialized_payload):
  if write_state.low is _LowWrite.OPEN:
    call.write(serialized_payload, operation_id)
    outstanding.add(_low.Event.Kind.WRITE_ACCEPTED)
    write_state.low = _LowWrite.ACTIVE
  elif write_state.low is _LowWrite.ACTIVE:
    write_state.pending.append(serialized_payload)
  else:
    raise ValueError('Write attempted after writes completed!')


class RearLink(base_interfaces.RearLink, activated.Activated):
  """An invocation-side bridge between RPC Framework and the C-ish _low code."""

  def __init__(
      self, host, port, pool, request_serializers, response_deserializers,
      secure, root_certificates, private_key, certificate_chain,
      metadata_transformer=None, server_host_override=None):
    """Constructor.

    Args:
      host: The host to which to connect for RPC service.
      port: The port to which to connect for RPC service.
      pool: A thread pool.
      request_serializers: A dict from RPC method names to request object
        serializer behaviors.
      response_deserializers: A dict from RPC method names to response object
        deserializer behaviors.
      secure: A boolean indicating whether or not to use a secure connection.
      root_certificates: The PEM-encoded root certificates or None to ask for
        them to be retrieved from a default location.
      private_key: The PEM-encoded private key to use or None if no private
        key should be used.
      certificate_chain: The PEM-encoded certificate chain to use or None if
        no certificate chain should be used.
      metadata_transformer: A function that given a metadata object produces
        another metadata to be used in the underlying communication on the
        wire.
      server_host_override: (For testing only) the target name used for SSL
        host name checking.
    """
    self._condition = threading.Condition()
    self._host = host
    self._port = port
    self._pool = pool
    self._request_serializers = request_serializers
    self._response_deserializers = response_deserializers

    self._fore_link = null.NULL_FORE_LINK
    self._completion_queue = None
    self._channel = None
    self._rpc_states = {}
    self._spinning = False
    if secure:
      self._client_credentials = _low.ClientCredentials(
          root_certificates, private_key, certificate_chain)
    else:
      self._client_credentials = None
    self._root_certificates = root_certificates
    self._private_key = private_key
    self._certificate_chain = certificate_chain
    self._metadata_transformer = metadata_transformer
    self._server_host_override = server_host_override

  def _on_write_event(self, operation_id, event, rpc_state):
    if event.write_accepted:
      if rpc_state.common.write.pending:
        rpc_state.call.write(
            rpc_state.common.write.pending.pop(0), operation_id)
        rpc_state.outstanding.add(_low.Event.Kind.WRITE_ACCEPTED)
      elif rpc_state.common.write.high is _common.HighWrite.CLOSED:
        rpc_state.call.complete(operation_id)
        rpc_state.outstanding.add(_low.Event.Kind.COMPLETE_ACCEPTED)
        rpc_state.common.write.low = _LowWrite.CLOSED
      else:
        rpc_state.common.write.low = _LowWrite.OPEN
    else:
      logging.error('RPC write not accepted! Event: %s', (event,))
      rpc_state.active = False
      ticket = base_interfaces.BackToFrontTicket(
          operation_id, rpc_state.common.sequence_number,
          base_interfaces.BackToFrontTicket.Kind.TRANSMISSION_FAILURE, None)
      rpc_state.common.sequence_number += 1
      self._fore_link.accept_back_to_front_ticket(ticket)

  def _on_read_event(self, operation_id, event, rpc_state):
    if event.bytes is not None:
      rpc_state.call.read(operation_id)
      rpc_state.outstanding.add(_low.Event.Kind.READ_ACCEPTED)

      ticket = base_interfaces.BackToFrontTicket(
          operation_id, rpc_state.common.sequence_number,
          base_interfaces.BackToFrontTicket.Kind.CONTINUATION,
          rpc_state.common.deserializer(event.bytes))
      rpc_state.common.sequence_number += 1
      self._fore_link.accept_back_to_front_ticket(ticket)

  def _on_complete_event(self, operation_id, event, rpc_state):
    if not event.complete_accepted:
      logging.error('RPC complete not accepted! Event: %s', (event,))
      rpc_state.active = False
      ticket = base_interfaces.BackToFrontTicket(
          operation_id, rpc_state.common.sequence_number,
          base_interfaces.BackToFrontTicket.Kind.TRANSMISSION_FAILURE, None)
      rpc_state.common.sequence_number += 1
      self._fore_link.accept_back_to_front_ticket(ticket)

  # TODO(nathaniel): Metadata support.
  def _on_metadata_event(self, operation_id, event, rpc_state):  # pylint: disable=unused-argument
    rpc_state.call.read(operation_id)
    rpc_state.outstanding.add(_low.Event.Kind.READ_ACCEPTED)

  def _on_finish_event(self, operation_id, event, rpc_state):
    """Handle termination of an RPC."""
    # TODO(nathaniel): Cover all statuses.
    if event.status.code is _low.Code.OK:
      kind = base_interfaces.BackToFrontTicket.Kind.COMPLETION
    elif event.status.code is _low.Code.CANCELLED:
      kind = base_interfaces.BackToFrontTicket.Kind.CANCELLATION
    elif event.status.code is _low.Code.DEADLINE_EXCEEDED:
      kind = base_interfaces.BackToFrontTicket.Kind.EXPIRATION
    else:
      kind = base_interfaces.BackToFrontTicket.Kind.TRANSMISSION_FAILURE
    ticket = base_interfaces.BackToFrontTicket(
        operation_id, rpc_state.common.sequence_number, kind, None)
    rpc_state.common.sequence_number += 1
    self._fore_link.accept_back_to_front_ticket(ticket)

  def _spin(self, completion_queue):
    while True:
      event = completion_queue.get(None)
      operation_id = event.tag

      with self._condition:
        rpc_state = self._rpc_states[operation_id]
        rpc_state.outstanding.remove(event.kind)
        if rpc_state.active and self._completion_queue is not None:
          if event.kind is _low.Event.Kind.WRITE_ACCEPTED:
            self._on_write_event(operation_id, event, rpc_state)
          elif event.kind is _low.Event.Kind.METADATA_ACCEPTED:
            self._on_metadata_event(operation_id, event, rpc_state)
          elif event.kind is _low.Event.Kind.READ_ACCEPTED:
            self._on_read_event(operation_id, event, rpc_state)
          elif event.kind is _low.Event.Kind.COMPLETE_ACCEPTED:
            self._on_complete_event(operation_id, event, rpc_state)
          elif event.kind is _low.Event.Kind.FINISH:
            self._on_finish_event(operation_id, event, rpc_state)
          else:
            logging.error('Illegal RPC event! %s', (event,))

        if not rpc_state.outstanding:
          self._rpc_states.pop(operation_id)
        if not self._rpc_states:
          self._spinning = False
          self._condition.notify_all()
          return

  def _invoke(self, operation_id, name, high_state, payload, timeout):
    """Invoke an RPC.

    Args:
      operation_id: Any object to be used as an operation ID for the RPC.
      name: The RPC method name.
      high_state: A _common.HighWrite value representing the "high write state"
        of the RPC.
      payload: A payload object for the RPC or None if no payload was given at
        invocation-time.
      timeout: A duration of time in seconds to allow for the RPC.
    """
    request_serializer = self._request_serializers[name]
    call = _low.Call(self._channel, self._completion_queue, name, self._host, time.time() + timeout)
    if self._metadata_transformer is not None:
      metadata = self._metadata_transformer([])
      for metadata_key, metadata_value in metadata:
        call.add_metadata(metadata_key, metadata_value)
    call.invoke(self._completion_queue, operation_id, operation_id)
    outstanding = set(_INVOCATION_EVENT_KINDS)

    if payload is None:
      if high_state is _common.HighWrite.CLOSED:
        call.complete(operation_id)
        low_state = _LowWrite.CLOSED
        outstanding.add(_low.Event.Kind.COMPLETE_ACCEPTED)
      else:
        low_state = _LowWrite.OPEN
    else:
      serialized_payload = request_serializer(payload)
      call.write(serialized_payload, operation_id)
      outstanding.add(_low.Event.Kind.WRITE_ACCEPTED)
      low_state = _LowWrite.ACTIVE

    write_state = _common.WriteState(low_state, high_state, [])
    common_state = _common.CommonRPCState(
        write_state, 0, self._response_deserializers[name], request_serializer)
    self._rpc_states[operation_id] = _RPCState(
        call, outstanding, True, common_state)

    if not self._spinning:
      self._pool.submit(self._spin, self._completion_queue)
      self._spinning = True

  def _commence(self, operation_id, name, payload, timeout):
    self._invoke(operation_id, name, _common.HighWrite.OPEN, payload, timeout)

  def _continue(self, operation_id, payload):
    rpc_state = self._rpc_states.get(operation_id, None)
    if rpc_state is None or not rpc_state.active:
      return

    _write(
        operation_id, rpc_state.call, rpc_state.outstanding,
        rpc_state.common.write, rpc_state.common.serializer(payload))

  def _complete(self, operation_id, payload):
    """Close writes associated with an ongoing RPC.

    Args:
      operation_id: Any object being use as an operation ID for the RPC.
      payload: A payload object for the RPC (and thus the last payload object
        for the RPC) or None if no payload was given along with the instruction
        to indicate the end of writes for the RPC.
    """
    rpc_state = self._rpc_states.get(operation_id, None)
    if rpc_state is None or not rpc_state.active:
      return

    write_state = rpc_state.common.write
    if payload is None:
      if write_state.low is _LowWrite.OPEN:
        rpc_state.call.complete(operation_id)
        rpc_state.outstanding.add(_low.Event.Kind.COMPLETE_ACCEPTED)
        write_state.low = _LowWrite.CLOSED
    else:
      _write(
          operation_id, rpc_state.call, rpc_state.outstanding, write_state,
          rpc_state.common.serializer(payload))
    write_state.high = _common.HighWrite.CLOSED

  def _entire(self, operation_id, name, payload, timeout):
    self._invoke(operation_id, name, _common.HighWrite.CLOSED, payload, timeout)

  def _cancel(self, operation_id):
    rpc_state = self._rpc_states.get(operation_id, None)
    if rpc_state is not None and rpc_state.active:
      rpc_state.call.cancel()
      rpc_state.active = False

  def join_fore_link(self, fore_link):
    """See base_interfaces.RearLink.join_fore_link for specification."""
    with self._condition:
      self._fore_link = null.NULL_FORE_LINK if fore_link is None else fore_link

  def _start(self):
    """Starts this RearLink.

    This method must be called before attempting to exchange tickets with this
    object.
    """
    with self._condition:
      self._completion_queue = _low.CompletionQueue()
      self._channel = _low.Channel(
          '%s:%d' % (self._host, self._port), self._client_credentials,
          server_host_override=self._server_host_override)
    return self

  def _stop(self):
    """Stops this RearLink.

    This method must be called for proper termination of this object, and no
    attempts to exchange tickets with this object may be made after this method
    has been called.
    """
    with self._condition:
      self._completion_queue.stop()
      self._completion_queue = None

      while self._spinning:
        self._condition.wait()

  def __enter__(self):
    """See activated.Activated.__enter__ for specification."""
    return self._start()

  def __exit__(self, exc_type, exc_val, exc_tb):
    """See activated.Activated.__exit__ for specification."""
    self._stop()
    return False

  def start(self):
    """See activated.Activated.start for specification."""
    return self._start()

  def stop(self):
    """See activated.Activated.stop for specification."""
    self._stop()

  def accept_front_to_back_ticket(self, ticket):
    """See base_interfaces.RearLink.accept_front_to_back_ticket for spec."""
    with self._condition:
      if self._completion_queue is None:
        return

      if ticket.kind is base_interfaces.FrontToBackTicket.Kind.COMMENCEMENT:
        self._commence(
            ticket.operation_id, ticket.name, ticket.payload, ticket.timeout)
      elif ticket.kind is base_interfaces.FrontToBackTicket.Kind.CONTINUATION:
        self._continue(ticket.operation_id, ticket.payload)
      elif ticket.kind is base_interfaces.FrontToBackTicket.Kind.COMPLETION:
        self._complete(ticket.operation_id, ticket.payload)
      elif ticket.kind is base_interfaces.FrontToBackTicket.Kind.ENTIRE:
        self._entire(
            ticket.operation_id, ticket.name, ticket.payload, ticket.timeout)
      elif ticket.kind is base_interfaces.FrontToBackTicket.Kind.CANCELLATION:
        self._cancel(ticket.operation_id)
      else:
        # NOTE(nathaniel): All other categories are treated as cancellation.
        self._cancel(ticket.operation_id)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Behaviors for servicing RPCs."""

# base_interfaces and interfaces are referenced from specification in this
# module.
from grpc.framework.base import interfaces as base_interfaces  # pylint: disable=unused-import
from grpc.framework.face import _control
from grpc.framework.face import exceptions
from grpc.framework.face import interfaces  # pylint: disable=unused-import
from grpc.framework.foundation import abandonment
from grpc.framework.foundation import callable_util
from grpc.framework.foundation import stream
from grpc.framework.foundation import stream_util


class _ValueInStreamOutConsumer(stream.Consumer):
  """A stream.Consumer that maps inputs one-to-many onto outputs."""

  def __init__(self, behavior, context, downstream):
    """Constructor.

    Args:
      behavior: A callable that takes a single value and an
        interfaces.RpcContext and returns a generator of arbitrarily many
        values.
      context: An interfaces.RpcContext.
      downstream: A stream.Consumer to which to pass the values generated by the
        given behavior.
    """
    self._behavior = behavior
    self._context = context
    self._downstream = downstream

  def consume(self, value):
    _control.pipe_iterator_to_consumer(
        self._behavior(value, self._context), self._downstream,
        self._context.is_active, False)

  def terminate(self):
    self._downstream.terminate()

  def consume_and_terminate(self, value):
    _control.pipe_iterator_to_consumer(
        self._behavior(value, self._context), self._downstream,
        self._context.is_active, True)


def _pool_wrap(behavior, operation_context):
  """Wraps an operation-related behavior so that it may be called in a pool.

  Args:
    behavior: A callable related to carrying out an operation.
    operation_context: A base_interfaces.OperationContext for the operation.

  Returns:
    A callable that when called carries out the behavior of the given callable
      and handles whatever exceptions it raises appropriately.
  """
  def translation(*args):
    try:
      behavior(*args)
    except (
        abandonment.Abandoned,
        exceptions.ExpirationError,
        exceptions.CancellationError,
        exceptions.ServicedError,
        exceptions.NetworkError) as e:
      if operation_context.is_active():
        operation_context.fail(e)
    except Exception as e:
      operation_context.fail(e)
  return callable_util.with_exceptions_logged(
      translation, _control.INTERNAL_ERROR_LOG_MESSAGE)


def adapt_inline_value_in_value_out(method):
  def adaptation(response_consumer, operation_context):
    rpc_context = _control.RpcContext(operation_context)
    return stream_util.TransformingConsumer(
        lambda request: method(request, rpc_context), response_consumer)
  return adaptation


def adapt_inline_value_in_stream_out(method):
  def adaptation(response_consumer, operation_context):
    rpc_context = _control.RpcContext(operation_context)
    return _ValueInStreamOutConsumer(method, rpc_context, response_consumer)
  return adaptation


def adapt_inline_stream_in_value_out(method, pool):
  def adaptation(response_consumer, operation_context):
    rendezvous = _control.Rendezvous()
    operation_context.add_termination_callback(rendezvous.set_outcome)
    def in_pool_thread():
      response_consumer.consume_and_terminate(
          method(rendezvous, _control.RpcContext(operation_context)))
    pool.submit(_pool_wrap(in_pool_thread, operation_context))
    return rendezvous
  return adaptation


def adapt_inline_stream_in_stream_out(method, pool):
  """Adapts an interfaces.InlineStreamInStreamOutMethod for use with Consumers.

   RPCs may be serviced by calling the return value of this function, passing
   request values to the stream.Consumer returned from that call, and receiving
   response values from the stream.Consumer passed to that call.

  Args:
    method: An interfaces.InlineStreamInStreamOutMethod.
    pool: A thread pool.

  Returns:
    A callable that takes a stream.Consumer and a
      base_interfaces.OperationContext and returns a stream.Consumer.
  """
  def adaptation(response_consumer, operation_context):
    rendezvous = _control.Rendezvous()
    operation_context.add_termination_callback(rendezvous.set_outcome)
    def in_pool_thread():
      _control.pipe_iterator_to_consumer(
          method(rendezvous, _control.RpcContext(operation_context)),
          response_consumer, operation_context.is_active, True)
    pool.submit(_pool_wrap(in_pool_thread, operation_context))
    return rendezvous
  return adaptation


def adapt_event_value_in_value_out(method):
  def adaptation(response_consumer, operation_context):
    def on_payload(payload):
      method(
          payload, response_consumer.consume_and_terminate,
          _control.RpcContext(operation_context))
    return _control.UnaryConsumer(on_payload)
  return adaptation


def adapt_event_value_in_stream_out(method):
  def adaptation(response_consumer, operation_context):
    def on_payload(payload):
      method(
          payload, response_consumer, _control.RpcContext(operation_context))
    return _control.UnaryConsumer(on_payload)
  return adaptation


def adapt_event_stream_in_value_out(method):
  def adaptation(response_consumer, operation_context):
    rpc_context = _control.RpcContext(operation_context)
    return method(response_consumer.consume_and_terminate, rpc_context)
  return adaptation


def adapt_event_stream_in_stream_out(method):
  def adaptation(response_consumer, operation_context):
    return method(response_consumer, _control.RpcContext(operation_context))
  return adaptation



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""The RPC-service-side bridge between RPC Framework and GRPC-on-the-wire."""

import enum
import logging
import threading
import time

from grpc._adapter import _common
from grpc._adapter import _intermediary_low as _low
from grpc.framework.base import interfaces as base_interfaces
from grpc.framework.base import null
from grpc.framework.foundation import activated
from grpc.framework.foundation import logging_pool

_THREAD_POOL_SIZE = 10


@enum.unique
class _LowWrite(enum.Enum):
  """The possible categories of low-level write state."""

  OPEN = 'OPEN'
  ACTIVE = 'ACTIVE'
  CLOSED = 'CLOSED'


def _write(call, rpc_state, payload):
  serialized_payload = rpc_state.serializer(payload)
  if rpc_state.write.low is _LowWrite.OPEN:
    call.write(serialized_payload, call)
    rpc_state.write.low = _LowWrite.ACTIVE
  else:
    rpc_state.write.pending.append(serialized_payload)


def _status(call, rpc_state):
  call.status(_low.Status(_low.Code.OK, ''), call)
  rpc_state.write.low = _LowWrite.CLOSED


class ForeLink(base_interfaces.ForeLink, activated.Activated):
  """A service-side bridge between RPC Framework and the C-ish _low code."""

  def __init__(
      self, pool, request_deserializers, response_serializers,
      root_certificates, key_chain_pairs, port=None):
    """Constructor.

    Args:
      pool: A thread pool.
      request_deserializers: A dict from RPC method names to request object
        deserializer behaviors.
      response_serializers: A dict from RPC method names to response object
        serializer behaviors.
      root_certificates: The PEM-encoded client root certificates as a
        bytestring or None.
      key_chain_pairs: A sequence of PEM-encoded private key-certificate chain
        pairs.
      port: The port on which to serve, or None to have a port selected
        automatically.
    """
    self._condition = threading.Condition()
    self._pool = pool
    self._request_deserializers = request_deserializers
    self._response_serializers = response_serializers
    self._root_certificates = root_certificates
    self._key_chain_pairs = key_chain_pairs
    self._requested_port = port

    self._rear_link = null.NULL_REAR_LINK
    self._completion_queue = None
    self._server = None
    self._rpc_states = {}
    self._spinning = False
    self._port = None

  def _on_stop_event(self):
    self._spinning = False
    self._condition.notify_all()

  def _on_service_acceptance_event(self, event, server):
    """Handle a service invocation event."""
    service_acceptance = event.service_acceptance
    if service_acceptance is None:
      return

    call = service_acceptance.call
    call.accept(self._completion_queue, call)
    # TODO(nathaniel): Metadata support.
    call.premetadata()
    call.read(call)
    method = service_acceptance.method

    self._rpc_states[call] = _common.CommonRPCState(
        _common.WriteState(_LowWrite.OPEN, _common.HighWrite.OPEN, []), 1,
        self._request_deserializers[method],
        self._response_serializers[method])

    ticket = base_interfaces.FrontToBackTicket(
        call, 0, base_interfaces.FrontToBackTicket.Kind.COMMENCEMENT, method,
        base_interfaces.ServicedSubscription.Kind.FULL, None, None,
        service_acceptance.deadline - time.time())
    self._rear_link.accept_front_to_back_ticket(ticket)

    server.service(None)

  def _on_read_event(self, event):
    """Handle data arriving during an RPC."""
    call = event.tag
    rpc_state = self._rpc_states.get(call, None)
    if rpc_state is None:
      return

    sequence_number = rpc_state.sequence_number
    rpc_state.sequence_number += 1
    if event.bytes is None:
      ticket = base_interfaces.FrontToBackTicket(
          call, sequence_number,
          base_interfaces.FrontToBackTicket.Kind.COMPLETION, None, None, None,
          None, None)
    else:
      call.read(call)
      ticket = base_interfaces.FrontToBackTicket(
          call, sequence_number,
          base_interfaces.FrontToBackTicket.Kind.CONTINUATION, None, None,
          None, rpc_state.deserializer(event.bytes), None)

    self._rear_link.accept_front_to_back_ticket(ticket)

  def _on_write_event(self, event):
    call = event.tag
    rpc_state = self._rpc_states.get(call, None)
    if rpc_state is None:
      return

    if rpc_state.write.pending:
      serialized_payload = rpc_state.write.pending.pop(0)
      call.write(serialized_payload, call)
    elif rpc_state.write.high is _common.HighWrite.CLOSED:
      _status(call, rpc_state)
    else:
      rpc_state.write.low = _LowWrite.OPEN

  def _on_complete_event(self, event):
    if not event.complete_accepted:
      logging.error('Complete not accepted! %s', (event,))
      call = event.tag
      rpc_state = self._rpc_states.pop(call, None)
      if rpc_state is None:
        return

      sequence_number = rpc_state.sequence_number
      rpc_state.sequence_number += 1
      ticket = base_interfaces.FrontToBackTicket(
          call, sequence_number,
          base_interfaces.FrontToBackTicket.Kind.TRANSMISSION_FAILURE, None,
          None, None, None, None)
      self._rear_link.accept_front_to_back_ticket(ticket)

  def _on_finish_event(self, event):
    """Handle termination of an RPC."""
    call = event.tag
    rpc_state = self._rpc_states.pop(call, None)
    if rpc_state is None:
      return

    code = event.status.code
    if code is _low.Code.OK:
      return

    sequence_number = rpc_state.sequence_number
    rpc_state.sequence_number += 1
    if code is _low.Code.CANCELLED:
      ticket = base_interfaces.FrontToBackTicket(
          call, sequence_number,
          base_interfaces.FrontToBackTicket.Kind.CANCELLATION, None, None,
          None, None, None)
    elif code is _low.Code.DEADLINE_EXCEEDED:
      ticket = base_interfaces.FrontToBackTicket(
          call, sequence_number,
          base_interfaces.FrontToBackTicket.Kind.EXPIRATION, None, None, None,
          None, None)
    else:
      # TODO(nathaniel): Better mapping of codes to ticket-categories
      ticket = base_interfaces.FrontToBackTicket(
          call, sequence_number,
          base_interfaces.FrontToBackTicket.Kind.TRANSMISSION_FAILURE, None,
          None, None, None, None)
    self._rear_link.accept_front_to_back_ticket(ticket)

  def _spin(self, completion_queue, server):
    while True:
      event = completion_queue.get(None)

      with self._condition:
        if event.kind is _low.Event.Kind.STOP:
          self._on_stop_event()
          return
        elif self._server is None:
          continue
        elif event.kind is _low.Event.Kind.SERVICE_ACCEPTED:
          self._on_service_acceptance_event(event, server)
        elif event.kind is _low.Event.Kind.READ_ACCEPTED:
          self._on_read_event(event)
        elif event.kind is _low.Event.Kind.WRITE_ACCEPTED:
          self._on_write_event(event)
        elif event.kind is _low.Event.Kind.COMPLETE_ACCEPTED:
          self._on_complete_event(event)
        elif event.kind is _low.Event.Kind.FINISH:
          self._on_finish_event(event)
        else:
          logging.error('Illegal event! %s', (event,))

  def _continue(self, call, payload):
    rpc_state = self._rpc_states.get(call, None)
    if rpc_state is None:
      return

    _write(call, rpc_state, payload)

  def _complete(self, call, payload):
    """Handle completion of the writes of an RPC."""
    rpc_state = self._rpc_states.get(call, None)
    if rpc_state is None:
      return

    if rpc_state.write.low is _LowWrite.OPEN:
      if payload is None:
        _status(call, rpc_state)
      else:
        _write(call, rpc_state, payload)
    elif rpc_state.write.low is _LowWrite.ACTIVE:
      if payload is not None:
        rpc_state.write.pending.append(rpc_state.serializer(payload))
    else:
      raise ValueError('Called to complete after having already completed!')
    rpc_state.write.high = _common.HighWrite.CLOSED

  def _cancel(self, call):
    call.cancel()
    self._rpc_states.pop(call, None)

  def join_rear_link(self, rear_link):
    """See base_interfaces.ForeLink.join_rear_link for specification."""
    self._rear_link = null.NULL_REAR_LINK if rear_link is None else rear_link

  def _start(self):
    """Starts this ForeLink.

    This method must be called before attempting to exchange tickets with this
    object.
    """
    with self._condition:
      address = '[::]:%d' % (
          0 if self._requested_port is None else self._requested_port)
      self._completion_queue = _low.CompletionQueue()
      if self._root_certificates is None and not self._key_chain_pairs:
        self._server = _low.Server(self._completion_queue)
        self._port = self._server.add_http2_addr(address)
      else:
        server_credentials = _low.ServerCredentials(
          self._root_certificates, self._key_chain_pairs)
        self._server = _low.Server(self._completion_queue)
        self._port = self._server.add_secure_http2_addr(
            address, server_credentials)
      self._server.start()

      self._server.service(None)

      self._pool.submit(self._spin, self._completion_queue, self._server)
      self._spinning = True

      return self

  # TODO(nathaniel): Expose graceful-shutdown semantics in which this object
  # enters a state in which it finishes ongoing RPCs but refuses new ones.
  def _stop(self):
    """Stops this ForeLink.

    This method must be called for proper termination of this object, and no
    attempts to exchange tickets with this object may be made after this method
    has been called.
    """
    with self._condition:
      self._server.stop()
      # TODO(nathaniel): Yep, this is weird. Deleting a server shouldn't have a
      # behaviorally significant side-effect.
      self._server = None
      self._completion_queue.stop()

      while self._spinning:
        self._condition.wait()

      self._port = None

  def __enter__(self):
    """See activated.Activated.__enter__ for specification."""
    return self._start()

  def __exit__(self, exc_type, exc_val, exc_tb):
    """See activated.Activated.__exit__ for specification."""
    self._stop()
    return False

  def start(self):
    """See activated.Activated.start for specification."""
    return self._start()

  def stop(self):
    """See activated.Activated.stop for specification."""
    self._stop()

  def port(self):
    """Identifies the port on which this ForeLink is servicing RPCs.

    Returns:
      The number of the port on which this ForeLink is servicing RPCs, or None
        if this ForeLink is not currently activated and servicing RPCs.
    """
    with self._condition:
      return self._port

  def accept_back_to_front_ticket(self, ticket):
    """See base_interfaces.ForeLink.accept_back_to_front_ticket for spec."""
    with self._condition:
      if self._server is None:
        return

      if ticket.kind is base_interfaces.BackToFrontTicket.Kind.CONTINUATION:
        self._continue(ticket.operation_id, ticket.payload)
      elif ticket.kind is base_interfaces.BackToFrontTicket.Kind.COMPLETION:
        self._complete(ticket.operation_id, ticket.payload)
      else:
        self._cancel(ticket.operation_id)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""State and behavior for ticket reception."""

import abc

from grpc.framework.base import interfaces
from grpc.framework.base import _interfaces

_INITIAL_FRONT_TO_BACK_TICKET_KINDS = (
    interfaces.FrontToBackTicket.Kind.COMMENCEMENT,
    interfaces.FrontToBackTicket.Kind.ENTIRE,
)


class _Receiver(object):
  """Common specification of different ticket-handling behavior."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def abort_if_abortive(self, ticket):
    """Aborts the operation if the ticket is abortive.

    Args:
      ticket: A just-arrived ticket.

    Returns:
      A boolean indicating whether or not this Receiver aborted the operation
        based on the ticket.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def receive(self, ticket):
    """Handles a just-arrived ticket.

    Args:
      ticket: A just-arrived ticket.

    Returns:
      A boolean indicating whether or not the ticket was terminal (i.e. whether
        or not non-abortive tickets are legal after this one).
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def reception_failure(self):
    """Aborts the operation with an indication of reception failure."""
    raise NotImplementedError()


def _abort(
    outcome, termination_manager, transmission_manager, ingestion_manager,
    expiration_manager):
  """Indicates abortion with the given outcome to the given managers."""
  termination_manager.abort(outcome)
  transmission_manager.abort(outcome)
  ingestion_manager.abort()
  expiration_manager.abort()


def _abort_if_abortive(
    ticket, abortive, termination_manager, transmission_manager,
    ingestion_manager, expiration_manager):
  """Determines a ticket's being abortive and if so aborts the operation.

  Args:
    ticket: A just-arrived ticket.
    abortive: A callable that takes a ticket and returns an interfaces.Outcome
      indicating that the operation should be aborted or None indicating that
      the operation should not be aborted.
    termination_manager: The operation's _interfaces.TerminationManager.
    transmission_manager: The operation's _interfaces.TransmissionManager.
    ingestion_manager: The operation's _interfaces.IngestionManager.
    expiration_manager: The operation's _interfaces.ExpirationManager.

  Returns:
    True if the operation was aborted; False otherwise.
  """
  abortion_outcome = abortive(ticket)
  if abortion_outcome is None:
    return False
  else:
    _abort(
        abortion_outcome, termination_manager, transmission_manager,
        ingestion_manager, expiration_manager)
    return True


def _reception_failure(
    termination_manager, transmission_manager, ingestion_manager,
    expiration_manager):
  """Aborts the operation with an indication of reception failure."""
  _abort(
      interfaces.Outcome.RECEPTION_FAILURE, termination_manager,
      transmission_manager, ingestion_manager, expiration_manager)


class _BackReceiver(_Receiver):
  """Ticket-handling specific to the back side of an operation."""

  def __init__(
      self, termination_manager, transmission_manager, ingestion_manager,
      expiration_manager):
    """Constructor.

    Args:
      termination_manager: The operation's _interfaces.TerminationManager.
      transmission_manager: The operation's _interfaces.TransmissionManager.
      ingestion_manager: The operation's _interfaces.IngestionManager.
      expiration_manager: The operation's _interfaces.ExpirationManager.
    """
    self._termination_manager = termination_manager
    self._transmission_manager = transmission_manager
    self._ingestion_manager = ingestion_manager
    self._expiration_manager = expiration_manager

    self._first_ticket_seen = False
    self._last_ticket_seen = False

  def _abortive(self, ticket):
    """Determines whether or not (and if so, how) a ticket is abortive.

    Args:
      ticket: A just-arrived ticket.

    Returns:
      An interfaces.Outcome value describing operation abortion if the
        ticket is abortive or None if the ticket is not abortive.
    """
    if ticket.kind is interfaces.FrontToBackTicket.Kind.CANCELLATION:
      return interfaces.Outcome.CANCELLED
    elif ticket.kind is interfaces.FrontToBackTicket.Kind.EXPIRATION:
      return interfaces.Outcome.EXPIRED
    elif ticket.kind is interfaces.FrontToBackTicket.Kind.SERVICED_FAILURE:
      return interfaces.Outcome.SERVICED_FAILURE
    elif ticket.kind is interfaces.FrontToBackTicket.Kind.RECEPTION_FAILURE:
      return interfaces.Outcome.SERVICED_FAILURE
    elif (ticket.kind in _INITIAL_FRONT_TO_BACK_TICKET_KINDS and
          self._first_ticket_seen):
      return interfaces.Outcome.RECEPTION_FAILURE
    elif self._last_ticket_seen:
      return interfaces.Outcome.RECEPTION_FAILURE
    else:
      return None

  def abort_if_abortive(self, ticket):
    """See _Receiver.abort_if_abortive for specification."""
    return _abort_if_abortive(
        ticket, self._abortive, self._termination_manager,
        self._transmission_manager, self._ingestion_manager,
        self._expiration_manager)

  def receive(self, ticket):
    """See _Receiver.receive for specification."""
    if ticket.timeout is not None:
      self._expiration_manager.change_timeout(ticket.timeout)

    if ticket.kind is interfaces.FrontToBackTicket.Kind.COMMENCEMENT:
      self._first_ticket_seen = True
      self._ingestion_manager.start(ticket.name)
      if ticket.payload is not None:
        self._ingestion_manager.consume(ticket.payload)
    elif ticket.kind is interfaces.FrontToBackTicket.Kind.CONTINUATION:
      self._ingestion_manager.consume(ticket.payload)
    elif ticket.kind is interfaces.FrontToBackTicket.Kind.COMPLETION:
      self._last_ticket_seen = True
      if ticket.payload is None:
        self._ingestion_manager.terminate()
      else:
        self._ingestion_manager.consume_and_terminate(ticket.payload)
    else:
      self._first_ticket_seen = True
      self._last_ticket_seen = True
      self._ingestion_manager.start(ticket.name)
      if ticket.payload is None:
        self._ingestion_manager.terminate()
      else:
        self._ingestion_manager.consume_and_terminate(ticket.payload)

  def reception_failure(self):
    """See _Receiver.reception_failure for specification."""
    _reception_failure(
        self._termination_manager, self._transmission_manager,
        self._ingestion_manager, self._expiration_manager)


class _FrontReceiver(_Receiver):
  """Ticket-handling specific to the front side of an operation."""

  def __init__(
      self, termination_manager, transmission_manager, ingestion_manager,
      expiration_manager):
    """Constructor.

    Args:
      termination_manager: The operation's _interfaces.TerminationManager.
      transmission_manager: The operation's _interfaces.TransmissionManager.
      ingestion_manager: The operation's _interfaces.IngestionManager.
      expiration_manager: The operation's _interfaces.ExpirationManager.
    """
    self._termination_manager = termination_manager
    self._transmission_manager = transmission_manager
    self._ingestion_manager = ingestion_manager
    self._expiration_manager = expiration_manager

    self._last_ticket_seen = False

  def _abortive(self, ticket):
    """Determines whether or not (and if so, how) a ticket is abortive.

    Args:
      ticket: A just-arrived ticket.

    Returns:
      An interfaces.Outcome value describing operation abortion if the ticket
        is abortive or None if the ticket is not abortive.
    """
    if ticket.kind is interfaces.BackToFrontTicket.Kind.CANCELLATION:
      return interfaces.Outcome.CANCELLED
    elif ticket.kind is interfaces.BackToFrontTicket.Kind.EXPIRATION:
      return interfaces.Outcome.EXPIRED
    elif ticket.kind is interfaces.BackToFrontTicket.Kind.SERVICER_FAILURE:
      return interfaces.Outcome.SERVICER_FAILURE
    elif ticket.kind is interfaces.BackToFrontTicket.Kind.RECEPTION_FAILURE:
      return interfaces.Outcome.SERVICER_FAILURE
    elif self._last_ticket_seen:
      return interfaces.Outcome.RECEPTION_FAILURE
    else:
      return None

  def abort_if_abortive(self, ticket):
    """See _Receiver.abort_if_abortive for specification."""
    return _abort_if_abortive(
        ticket, self._abortive, self._termination_manager,
        self._transmission_manager, self._ingestion_manager,
        self._expiration_manager)

  def receive(self, ticket):
    """See _Receiver.receive for specification."""
    if ticket.kind is interfaces.BackToFrontTicket.Kind.CONTINUATION:
      self._ingestion_manager.consume(ticket.payload)
    elif ticket.kind is interfaces.BackToFrontTicket.Kind.COMPLETION:
      self._last_ticket_seen = True
      if ticket.payload is None:
        self._ingestion_manager.terminate()
      else:
        self._ingestion_manager.consume_and_terminate(ticket.payload)

  def reception_failure(self):
    """See _Receiver.reception_failure for specification."""
    _reception_failure(
        self._termination_manager, self._transmission_manager,
        self._ingestion_manager, self._expiration_manager)


class _ReceptionManager(_interfaces.ReceptionManager):
  """A ReceptionManager based around a _Receiver passed to it."""

  def __init__(self, lock, receiver):
    """Constructor.

    Args:
      lock: The operation-servicing-wide lock object.
      receiver: A _Receiver responsible for handling received tickets.
    """
    self._lock = lock
    self._receiver = receiver

    self._lowest_unseen_sequence_number = 0
    self._out_of_sequence_tickets = {}
    self._completed_sequence_number = None
    self._aborted = False

  def _sequence_failure(self, ticket):
    """Determines a just-arrived ticket's sequential legitimacy.

    Args:
      ticket: A just-arrived ticket.

    Returns:
      True if the ticket is sequentially legitimate; False otherwise.
    """
    if ticket.sequence_number < self._lowest_unseen_sequence_number:
      return True
    elif ticket.sequence_number in self._out_of_sequence_tickets:
      return True
    elif (self._completed_sequence_number is not None and
          self._completed_sequence_number <= ticket.sequence_number):
      return True
    else:
      return False

  def _process(self, ticket):
    """Process those tickets ready to be processed.

    Args:
      ticket: A just-arrived ticket the sequence number of which matches this
        _ReceptionManager's _lowest_unseen_sequence_number field.
    """
    while True:
      completed = self._receiver.receive(ticket)
      if completed:
        self._out_of_sequence_tickets.clear()
        self._completed_sequence_number = ticket.sequence_number
        self._lowest_unseen_sequence_number = ticket.sequence_number + 1
        return
      else:
        next_ticket = self._out_of_sequence_tickets.pop(
            ticket.sequence_number + 1, None)
        if next_ticket is None:
          self._lowest_unseen_sequence_number = ticket.sequence_number + 1
          return
        else:
          ticket = next_ticket

  def receive_ticket(self, ticket):
    """See _interfaces.ReceptionManager.receive_ticket for specification."""
    with self._lock:
      if self._aborted:
        return
      elif self._sequence_failure(ticket):
        self._receiver.reception_failure()
        self._aborted = True
      elif self._receiver.abort_if_abortive(ticket):
        self._aborted = True
      elif ticket.sequence_number == self._lowest_unseen_sequence_number:
        self._process(ticket)
      else:
        self._out_of_sequence_tickets[ticket.sequence_number] = ticket


def front_reception_manager(
    lock, termination_manager, transmission_manager, ingestion_manager,
    expiration_manager):
  """Creates a _interfaces.ReceptionManager for front-side use.

  Args:
    lock: The operation-servicing-wide lock object.
    termination_manager: The operation's _interfaces.TerminationManager.
    transmission_manager: The operation's _interfaces.TransmissionManager.
    ingestion_manager: The operation's _interfaces.IngestionManager.
    expiration_manager: The operation's _interfaces.ExpirationManager.

  Returns:
    A _interfaces.ReceptionManager appropriate for front-side use.
  """
  return _ReceptionManager(
      lock, _FrontReceiver(
          termination_manager, transmission_manager, ingestion_manager,
          expiration_manager))


def back_reception_manager(
    lock, termination_manager, transmission_manager, ingestion_manager,
    expiration_manager):
  """Creates a _interfaces.ReceptionManager for back-side use.

  Args:
    lock: The operation-servicing-wide lock object.
    termination_manager: The operation's _interfaces.TerminationManager.
    transmission_manager: The operation's _interfaces.TransmissionManager.
    ingestion_manager: The operation's _interfaces.IngestionManager.
    expiration_manager: The operation's _interfaces.ExpirationManager.

  Returns:
    A _interfaces.ReceptionManager appropriate for back-side use.
  """
  return _ReceptionManager(
      lock, _BackReceiver(
          termination_manager, transmission_manager, ingestion_manager,
          expiration_manager))



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Temporary old _low-like layer.

Eases refactoring burden while we overhaul the Python framework.

Plan:
    The layers used to look like:
        ...                # outside _adapter
        fore.py + rear.py  # visible outside _adapter
        _low
        _c
    The layers currently look like:
        ...                # outside _adapter
        fore.py + rear.py  # visible outside _adapter
        _low_intermediary  # adapter for new '_low' to old '_low'
        _low               # new '_low'
        _c                 # new '_c'
    We will later remove _low_intermediary after refactoring of fore.py and
    rear.py according to the ticket system refactoring and get:
        ...                # outside _adapter, refactored
        fore.py + rear.py  # visible outside _adapter, refactored
        _low               # new '_low'
        _c                 # new '_c'
"""

import collections
import enum

from grpc._adapter import _low
from grpc._adapter import _types

_IGNORE_ME_TAG = object()
Code = _types.StatusCode


class Status(collections.namedtuple('Status', ['code', 'details'])):
  """Describes an RPC's overall status."""


class ServiceAcceptance(
    collections.namedtuple(
        'ServiceAcceptance', ['call', 'method', 'host', 'deadline'])):
  """Describes an RPC on the service side at the start of service."""


class Event(
    collections.namedtuple(
        'Event',
        ['kind', 'tag', 'write_accepted', 'complete_accepted',
         'service_acceptance', 'bytes', 'status', 'metadata'])):
  """Describes an event emitted from a completion queue."""

  @enum.unique
  class Kind(enum.Enum):
    """Describes the kind of an event."""

    STOP = object()
    WRITE_ACCEPTED = object()
    COMPLETE_ACCEPTED = object()
    SERVICE_ACCEPTED = object()
    READ_ACCEPTED = object()
    METADATA_ACCEPTED = object()
    FINISH = object()


class _TagAdapter(collections.namedtuple('_TagAdapter', [
      'user_tag',
      'kind'
  ])):
  pass


class Call(object):
  """Adapter from old _low.Call interface to new _low.Call."""

  def __init__(self, channel, completion_queue, method, host, deadline):
    self._internal = channel._internal.create_call(
        completion_queue._internal, method, host, deadline)
    self._metadata = []

  @staticmethod
  def _from_internal(internal):
    call = Call.__new__(Call)
    call._internal = internal
    call._metadata = []
    return call

  def invoke(self, completion_queue, metadata_tag, finish_tag):
    err0 = self._internal.start_batch([
          _types.OpArgs.send_initial_metadata(self._metadata)
      ], _IGNORE_ME_TAG)
    err1 = self._internal.start_batch([
          _types.OpArgs.recv_initial_metadata()
      ], _TagAdapter(metadata_tag, Event.Kind.METADATA_ACCEPTED))
    err2 = self._internal.start_batch([
          _types.OpArgs.recv_status_on_client()
      ], _TagAdapter(finish_tag, Event.Kind.FINISH))
    return err0 if err0 != _types.CallError.OK else err1 if err1 != _types.CallError.OK else err2 if err2 != _types.CallError.OK else _types.CallError.OK

  def write(self, message, tag):
    return self._internal.start_batch([
          _types.OpArgs.send_message(message)
      ], _TagAdapter(tag, Event.Kind.WRITE_ACCEPTED))

  def complete(self, tag):
    return self._internal.start_batch([
          _types.OpArgs.send_close_from_client()
      ], _TagAdapter(tag, Event.Kind.COMPLETE_ACCEPTED))

  def accept(self, completion_queue, tag):
    return self._internal.start_batch([
          _types.OpArgs.recv_close_on_server()
      ], _TagAdapter(tag, Event.Kind.FINISH))

  def add_metadata(self, key, value):
    self._metadata.append((key, value))

  def premetadata(self):
    result = self._internal.start_batch([
          _types.OpArgs.send_initial_metadata(self._metadata)
      ], _IGNORE_ME_TAG)
    self._metadata = []
    return result

  def read(self, tag):
    return self._internal.start_batch([
          _types.OpArgs.recv_message()
      ], _TagAdapter(tag, Event.Kind.READ_ACCEPTED))

  def status(self, status, tag):
    return self._internal.start_batch([
          _types.OpArgs.send_status_from_server(self._metadata, status.code, status.details)
      ], _TagAdapter(tag, Event.Kind.COMPLETE_ACCEPTED))

  def cancel(self):
    return self._internal.cancel()


class Channel(object):
  """Adapter from old _low.Channel interface to new _low.Channel."""

  def __init__(self, hostport, client_credentials, server_host_override=None):
    args = []
    if server_host_override:
      args.append((_types.GrpcChannelArgumentKeys.SSL_TARGET_NAME_OVERRIDE.value, server_host_override))
    creds = None
    if client_credentials:
      creds = client_credentials._internal
    self._internal = _low.Channel(hostport, args, creds)


class CompletionQueue(object):
  """Adapter from old _low.CompletionQueue interface to new _low.CompletionQueue."""

  def __init__(self):
    self._internal = _low.CompletionQueue()

  def get(self, deadline=None):
    if deadline is None:
      ev = self._internal.next()
    else:
      ev = self._internal.next(deadline)
    if ev is None:
      return None
    elif ev.tag is _IGNORE_ME_TAG:
      return self.get(deadline)
    elif ev.type == _types.EventType.QUEUE_SHUTDOWN:
      kind = Event.Kind.STOP
      tag = None
      write_accepted = None
      complete_accepted = None
      service_acceptance = None
      message_bytes = None
      status = None
      metadata = None
    elif ev.type == _types.EventType.OP_COMPLETE:
      kind = ev.tag.kind
      tag = ev.tag.user_tag
      write_accepted = ev.success if kind == Event.Kind.WRITE_ACCEPTED else None
      complete_accepted = ev.success if kind == Event.Kind.COMPLETE_ACCEPTED else None
      service_acceptance = ServiceAcceptance(Call._from_internal(ev.call), ev.call_details.method, ev.call_details.host, ev.call_details.deadline) if kind == Event.Kind.SERVICE_ACCEPTED else None
      message_bytes = ev.results[0].message if kind == Event.Kind.READ_ACCEPTED else None
      status = Status(ev.results[0].status.code, ev.results[0].status.details) if (kind == Event.Kind.FINISH and ev.results[0].status) else Status(_types.StatusCode.CANCELLED if ev.results[0].cancelled else _types.StatusCode.OK, '') if len(ev.results) > 0 and ev.results[0].cancelled is not None else None
      metadata = ev.results[0].initial_metadata if (kind in [Event.Kind.SERVICE_ACCEPTED, Event.Kind.METADATA_ACCEPTED]) else (ev.results[0].trailing_metadata if kind == Event.Kind.FINISH else None)
    else:
      raise RuntimeError('unknown event')
    result_ev = Event(kind=kind, tag=tag, write_accepted=write_accepted, complete_accepted=complete_accepted, service_acceptance=service_acceptance, bytes=message_bytes, status=status, metadata=metadata)
    return result_ev

  def stop(self):
    self._internal.shutdown()


class Server(object):
  """Adapter from old _low.Server interface to new _low.Server."""

  def __init__(self, completion_queue):
    self._internal = _low.Server(completion_queue._internal, [])
    self._internal_cq = completion_queue._internal

  def add_http2_addr(self, addr):
    return self._internal.add_http2_port(addr)

  def add_secure_http2_addr(self, addr, server_credentials):
    if server_credentials is None:
      return self._internal.add_http2_port(addr, None)
    else:
      return self._internal.add_http2_port(addr, server_credentials._internal)

  def start(self):
    return self._internal.start()

  def service(self, tag):
    return self._internal.request_call(self._internal_cq, _TagAdapter(tag, Event.Kind.SERVICE_ACCEPTED))

  def stop(self):
    return self._internal.shutdown(_TagAdapter(None, Event.Kind.STOP))


class ClientCredentials(object):
  """Adapter from old _low.ClientCredentials interface to new _low.ClientCredentials."""

  def __init__(self, root_certificates, private_key, certificate_chain):
    self._internal = _low.ClientCredentials.ssl(root_certificates, private_key, certificate_chain)


class ServerCredentials(object):
  """Adapter from old _low.ServerCredentials interface to new _low.ServerCredentials."""

  def __init__(self, root_credentials, pair_sequence):
    self._internal = _low.ServerCredentials.ssl(root_credentials, list(pair_sequence))



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from grpc.framework.common import cardinality
from grpc.framework.face import exceptions as face_exceptions
from grpc.framework.face import interfaces as face_interfaces
from grpc.framework.foundation import future
from grpc.framework.alpha import exceptions
from grpc.framework.alpha import interfaces

_EARLY_ADOPTER_CARDINALITY_TO_COMMON_CARDINALITY = {
    interfaces.Cardinality.UNARY_UNARY: cardinality.Cardinality.UNARY_UNARY,
    interfaces.Cardinality.UNARY_STREAM: cardinality.Cardinality.UNARY_STREAM,
    interfaces.Cardinality.STREAM_UNARY: cardinality.Cardinality.STREAM_UNARY,
    interfaces.Cardinality.STREAM_STREAM: cardinality.Cardinality.STREAM_STREAM,
}

_ABORTION_REEXPORT = {
    face_interfaces.Abortion.CANCELLED: interfaces.Abortion.CANCELLED,
    face_interfaces.Abortion.EXPIRED: interfaces.Abortion.EXPIRED,
    face_interfaces.Abortion.NETWORK_FAILURE:
        interfaces.Abortion.NETWORK_FAILURE,
    face_interfaces.Abortion.SERVICED_FAILURE:
        interfaces.Abortion.SERVICED_FAILURE,
    face_interfaces.Abortion.SERVICER_FAILURE:
        interfaces.Abortion.SERVICER_FAILURE,
}


class _RpcError(exceptions.RpcError):
  pass


def _reexport_error(face_rpc_error):
  if isinstance(face_rpc_error, face_exceptions.CancellationError):
    return exceptions.CancellationError()
  elif isinstance(face_rpc_error, face_exceptions.ExpirationError):
    return exceptions.ExpirationError()
  else:
    return _RpcError()


def _as_face_abortion_callback(abortion_callback):
  def face_abortion_callback(face_abortion):
    abortion_callback(_ABORTION_REEXPORT[face_abortion])
  return face_abortion_callback


class _ReexportedFuture(future.Future):

  def __init__(self, face_future):
    self._face_future = face_future

  def cancel(self):
    return self._face_future.cancel()

  def cancelled(self):
    return self._face_future.cancelled()

  def running(self):
    return self._face_future.running()

  def done(self):
    return self._face_future.done()

  def result(self, timeout=None):
    try:
      return self._face_future.result(timeout=timeout)
    except face_exceptions.RpcError as e:
      raise _reexport_error(e)

  def exception(self, timeout=None):
    face_error = self._face_future.exception(timeout=timeout)
    return None if face_error is None else _reexport_error(face_error)

  def traceback(self, timeout=None):
    return self._face_future.traceback(timeout=timeout)

  def add_done_callback(self, fn):
    self._face_future.add_done_callback(lambda unused_face_future: fn(self))


def _call_reexporting_errors(behavior, *args, **kwargs):
  try:
    return behavior(*args, **kwargs)
  except face_exceptions.RpcError as e:
    raise _reexport_error(e)


def _reexported_future(face_future):
  return _ReexportedFuture(face_future)


class _CancellableIterator(interfaces.CancellableIterator):

  def __init__(self, face_cancellable_iterator):
    self._face_cancellable_iterator = face_cancellable_iterator

  def __iter__(self):
    return self

  def next(self):
    return _call_reexporting_errors(self._face_cancellable_iterator.next)

  def cancel(self):
    self._face_cancellable_iterator.cancel()


class _RpcContext(interfaces.RpcContext):

  def __init__(self, face_rpc_context):
    self._face_rpc_context = face_rpc_context

  def is_active(self):
    return self._face_rpc_context.is_active()

  def time_remaining(self):
    return self._face_rpc_context.time_remaining()

  def add_abortion_callback(self, abortion_callback):
    self._face_rpc_context.add_abortion_callback(
        _as_face_abortion_callback(abortion_callback))


class _UnaryUnarySyncAsync(interfaces.UnaryUnarySyncAsync):

  def __init__(self, face_unary_unary_multi_callable):
    self._underlying = face_unary_unary_multi_callable

  def __call__(self, request, timeout):
    return _call_reexporting_errors(
        self._underlying, request, timeout)

  def async(self, request, timeout):
    return _ReexportedFuture(self._underlying.future(request, timeout))


class _StreamUnarySyncAsync(interfaces.StreamUnarySyncAsync):

  def __init__(self, face_stream_unary_multi_callable):
    self._underlying = face_stream_unary_multi_callable

  def __call__(self, request_iterator, timeout):
    return _call_reexporting_errors(
        self._underlying, request_iterator, timeout)

  def async(self, request_iterator, timeout):
    return _ReexportedFuture(self._underlying.future(request_iterator, timeout))


def common_cardinality(early_adopter_cardinality):
  return _EARLY_ADOPTER_CARDINALITY_TO_COMMON_CARDINALITY[
      early_adopter_cardinality]


def common_cardinalities(early_adopter_cardinalities):
  common_cardinalities = {}
  for name, early_adopter_cardinality in early_adopter_cardinalities.iteritems():
    common_cardinalities[name] = _EARLY_ADOPTER_CARDINALITY_TO_COMMON_CARDINALITY[
        early_adopter_cardinality]
  return common_cardinalities


def rpc_context(face_rpc_context):
  return _RpcContext(face_rpc_context)


def cancellable_iterator(face_cancellable_iterator):
  return _CancellableIterator(face_cancellable_iterator)


def unary_unary_sync_async(face_unary_unary_multi_callable):
  return _UnaryUnarySyncAsync(face_unary_unary_multi_callable)


def stream_unary_sync_async(face_stream_unary_multi_callable):
  return _StreamUnarySyncAsync(face_stream_unary_multi_callable)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import threading
import time
import unittest

from grpc._adapter import _types
from grpc._adapter import _low


def WaitForEvents(completion_queues, deadline):
  """
  Args:
    completion_queues: list of completion queues to wait for events on
    deadline: absolute deadline to wait until

  Returns:
    a sequence of events of length len(completion_queues).
  """

  results = [None] * len(completion_queues)
  lock = threading.Lock()
  threads = []
  def set_ith_result(i, completion_queue):
    result = completion_queue.next(deadline)
    with lock:
      print i, completion_queue, result, time.time() - deadline
      results[i] = result
  for i, completion_queue in enumerate(completion_queues):
    thread = threading.Thread(target=set_ith_result,
                              args=[i, completion_queue])
    thread.start()
    threads.append(thread)
  for thread in threads:
    thread.join()
  return results

class InsecureServerInsecureClient(unittest.TestCase):

  def setUp(self):
    self.server_completion_queue = _low.CompletionQueue()
    self.server = _low.Server(self.server_completion_queue, [])
    self.port = self.server.add_http2_port('[::]:0')
    self.client_completion_queue = _low.CompletionQueue()
    self.client_channel = _low.Channel('localhost:%d'%self.port, [])

    self.server.start()

  def tearDown(self):
    self.server.shutdown()
    del self.client_channel

    self.client_completion_queue.shutdown()
    while self.client_completion_queue.next().type != _types.EventType.QUEUE_SHUTDOWN:
      pass
    self.server_completion_queue.shutdown()
    while self.server_completion_queue.next().type != _types.EventType.QUEUE_SHUTDOWN:
      pass

    del self.client_completion_queue
    del self.server_completion_queue
    del self.server

  def testEcho(self):
    DEADLINE = time.time()+5
    DEADLINE_TOLERANCE = 0.25
    CLIENT_METADATA_ASCII_KEY = 'key'
    CLIENT_METADATA_ASCII_VALUE = 'val'
    CLIENT_METADATA_BIN_KEY = 'key-bin'
    CLIENT_METADATA_BIN_VALUE = b'\0'*1000
    SERVER_INITIAL_METADATA_KEY = 'init_me_me_me'
    SERVER_INITIAL_METADATA_VALUE = 'whodawha?'
    SERVER_TRAILING_METADATA_KEY = 'California_is_in_a_drought'
    SERVER_TRAILING_METADATA_VALUE = 'zomg it is'
    SERVER_STATUS_CODE = _types.StatusCode.OK
    SERVER_STATUS_DETAILS = 'our work is never over'
    REQUEST = 'in death a member of project mayhem has a name'
    RESPONSE = 'his name is robert paulson'
    METHOD = 'twinkies'
    HOST = 'hostess'
    server_request_tag = object()
    request_call_result = self.server.request_call(self.server_completion_queue, server_request_tag)

    self.assertEquals(_types.CallError.OK, request_call_result)

    client_call_tag = object()
    client_call = self.client_channel.create_call(self.client_completion_queue, METHOD, HOST, DEADLINE)
    client_initial_metadata = [(CLIENT_METADATA_ASCII_KEY, CLIENT_METADATA_ASCII_VALUE), (CLIENT_METADATA_BIN_KEY, CLIENT_METADATA_BIN_VALUE)]
    client_start_batch_result = client_call.start_batch([
        _types.OpArgs.send_initial_metadata(client_initial_metadata),
        _types.OpArgs.send_message(REQUEST),
        _types.OpArgs.send_close_from_client(),
        _types.OpArgs.recv_initial_metadata(),
        _types.OpArgs.recv_message(),
        _types.OpArgs.recv_status_on_client()
    ], client_call_tag)
    self.assertEquals(_types.CallError.OK, client_start_batch_result)

    client_no_event, request_event, = WaitForEvents([self.client_completion_queue, self.server_completion_queue], time.time() + 2)
    self.assertEquals(client_no_event, None)
    self.assertEquals(_types.EventType.OP_COMPLETE, request_event.type)
    self.assertIsInstance(request_event.call, _low.Call)
    self.assertIs(server_request_tag, request_event.tag)
    self.assertEquals(1, len(request_event.results))
    self.assertEquals(dict(client_initial_metadata), dict(request_event.results[0].initial_metadata))
    self.assertEquals(METHOD, request_event.call_details.method)
    self.assertEquals(HOST, request_event.call_details.host)
    self.assertLess(abs(DEADLINE - request_event.call_details.deadline), DEADLINE_TOLERANCE)

    server_call_tag = object()
    server_call = request_event.call
    server_initial_metadata = [(SERVER_INITIAL_METADATA_KEY, SERVER_INITIAL_METADATA_VALUE)]
    server_trailing_metadata = [(SERVER_TRAILING_METADATA_KEY, SERVER_TRAILING_METADATA_VALUE)]
    server_start_batch_result = server_call.start_batch([
        _types.OpArgs.send_initial_metadata(server_initial_metadata),
        _types.OpArgs.recv_message(),
        _types.OpArgs.send_message(RESPONSE),
        _types.OpArgs.recv_close_on_server(),
        _types.OpArgs.send_status_from_server(server_trailing_metadata, SERVER_STATUS_CODE, SERVER_STATUS_DETAILS)
    ], server_call_tag)
    self.assertEquals(_types.CallError.OK, server_start_batch_result)

    client_event, server_event, = WaitForEvents([self.client_completion_queue, self.server_completion_queue], time.time() + 1)

    self.assertEquals(6, len(client_event.results))
    found_client_op_types = set()
    for client_result in client_event.results:
      self.assertNotIn(client_result.type, found_client_op_types)  # we expect each op type to be unique
      found_client_op_types.add(client_result.type)
      if client_result.type == _types.OpType.RECV_INITIAL_METADATA:
        self.assertEquals(dict(server_initial_metadata), dict(client_result.initial_metadata))
      elif client_result.type == _types.OpType.RECV_MESSAGE:
        self.assertEquals(RESPONSE, client_result.message)
      elif client_result.type == _types.OpType.RECV_STATUS_ON_CLIENT:
        self.assertEquals(dict(server_trailing_metadata), dict(client_result.trailing_metadata))
        self.assertEquals(SERVER_STATUS_DETAILS, client_result.status.details)
        self.assertEquals(SERVER_STATUS_CODE, client_result.status.code)
    self.assertEquals(set([
          _types.OpType.SEND_INITIAL_METADATA,
          _types.OpType.SEND_MESSAGE,
          _types.OpType.SEND_CLOSE_FROM_CLIENT,
          _types.OpType.RECV_INITIAL_METADATA,
          _types.OpType.RECV_MESSAGE,
          _types.OpType.RECV_STATUS_ON_CLIENT
      ]), found_client_op_types)

    self.assertEquals(5, len(server_event.results))
    found_server_op_types = set()
    for server_result in server_event.results:
      self.assertNotIn(client_result.type, found_server_op_types)
      found_server_op_types.add(server_result.type)
      if server_result.type == _types.OpType.RECV_MESSAGE:
        self.assertEquals(REQUEST, server_result.message)
      elif server_result.type == _types.OpType.RECV_CLOSE_ON_SERVER:
        self.assertFalse(server_result.cancelled)
    self.assertEquals(set([
          _types.OpType.SEND_INITIAL_METADATA,
          _types.OpType.RECV_MESSAGE,
          _types.OpType.SEND_MESSAGE,
          _types.OpType.RECV_CLOSE_ON_SERVER,
          _types.OpType.SEND_STATUS_FROM_SERVER
      ]), found_server_op_types)

    del client_call
    del server_call


if __name__ == '__main__':
  unittest.main(verbosity=2)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Tests for grpc.framework.base.implementations."""

import unittest

from grpc.framework.base import implementations
from grpc.framework.base import interfaces_test_case
from grpc.framework.base import util
from grpc.framework.foundation import logging_pool

POOL_MAX_WORKERS = 10
DEFAULT_TIMEOUT = 30
MAXIMUM_TIMEOUT = 60


class ImplementationsTest(
    interfaces_test_case.FrontAndBackTest, unittest.TestCase):

  def setUp(self):
    self.memory_transmission_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.front_work_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.front_transmission_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.front_utility_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.back_work_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.back_transmission_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.back_utility_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.test_pool = logging_pool.pool(POOL_MAX_WORKERS)
    self.test_servicer = interfaces_test_case.TestServicer(self.test_pool)
    self.front = implementations.front_link(
        self.front_work_pool, self.front_transmission_pool,
        self.front_utility_pool)
    self.back = implementations.back_link(
        self.test_servicer, self.back_work_pool, self.back_transmission_pool,
        self.back_utility_pool, DEFAULT_TIMEOUT, MAXIMUM_TIMEOUT)
    self.front.join_rear_link(self.back)
    self.back.join_fore_link(self.front)

  def tearDown(self):
    util.wait_for_idle(self.back)
    util.wait_for_idle(self.front)
    self.memory_transmission_pool.shutdown(wait=True)
    self.front_work_pool.shutdown(wait=True)
    self.front_transmission_pool.shutdown(wait=True)
    self.front_utility_pool.shutdown(wait=True)
    self.back_work_pool.shutdown(wait=True)
    self.back_transmission_pool.shutdown(wait=True)
    self.back_utility_pool.shutdown(wait=True)
    self.test_pool.shutdown(wait=True)


if __name__ == '__main__':
  unittest.main(verbosity=2)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Interfaces for the face layer of RPC Framework."""

import abc
import enum

# cardinality, style, exceptions, abandonment, future, and stream are
# referenced from specification in this module.
from grpc.framework.common import cardinality  # pylint: disable=unused-import
from grpc.framework.common import style  # pylint: disable=unused-import
from grpc.framework.face import exceptions  # pylint: disable=unused-import
from grpc.framework.foundation import abandonment  # pylint: disable=unused-import
from grpc.framework.foundation import future  # pylint: disable=unused-import
from grpc.framework.foundation import stream  # pylint: disable=unused-import


@enum.unique
class Abortion(enum.Enum):
  """Categories of RPC abortion."""
  CANCELLED = 'cancelled'
  EXPIRED = 'expired'
  NETWORK_FAILURE = 'network failure'
  SERVICED_FAILURE = 'serviced failure'
  SERVICER_FAILURE = 'servicer failure'


class CancellableIterator(object):
  """Implements the Iterator protocol and affords a cancel method."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __iter__(self):
    """Returns the self object in accordance with the Iterator protocol."""
    raise NotImplementedError()

  @abc.abstractmethod
  def next(self):
    """Returns a value or raises StopIteration per the Iterator protocol."""
    raise NotImplementedError()

  @abc.abstractmethod
  def cancel(self):
    """Requests cancellation of whatever computation underlies this iterator."""
    raise NotImplementedError()


class RpcContext(object):
  """Provides RPC-related information and action."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def is_active(self):
    """Describes whether the RPC is active or has terminated."""
    raise NotImplementedError()

  @abc.abstractmethod
  def time_remaining(self):
    """Describes the length of allowed time remaining for the RPC.

    Returns:
      A nonnegative float indicating the length of allowed time in seconds
      remaining for the RPC to complete before it is considered to have timed
      out.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def add_abortion_callback(self, abortion_callback):
    """Registers a callback to be called if the RPC is aborted.

    Args:
      abortion_callback: A callable to be called and passed an Abortion value
        in the event of RPC abortion.
    """
    raise NotImplementedError()


class Call(object):
  """Invocation-side representation of an RPC.

  Attributes:
    context: An RpcContext affording information about the RPC.
  """
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def cancel(self):
    """Requests cancellation of the RPC."""
    raise NotImplementedError()


class UnaryUnaryMultiCallable(object):
  """Affords invoking a unary-unary RPC in any call style."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __call__(self, request, timeout):
    """Synchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      The response value for the RPC.

    Raises:
      exceptions.RpcError: Indicating that the RPC was aborted.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def future(self, request, timeout):
    """Asynchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A future.Future representing the RPC. In the event of RPC completion, the
        returned Future's result value will be the response value of the RPC.
        In the event of RPC abortion, the returned Future's exception value
        will be an exceptions.RpcError.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event(self, request, response_callback, abortion_callback, timeout):
    """Asynchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      response_callback: A callback to be called to accept the restponse value
        of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A Call object for the RPC.
    """
    raise NotImplementedError()


class UnaryStreamMultiCallable(object):
  """Affords invoking a unary-stream RPC in any call style."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __call__(self, request, timeout):
    """Synchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A CancellableIterator that yields the response values of the RPC and
        affords RPC cancellation. Drawing response values from the returned
        CancellableIterator may raise exceptions.RpcError indicating abortion
        of the RPC.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event(self, request, response_consumer, abortion_callback, timeout):
    """Asynchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      response_consumer: A stream.Consumer to be called to accept the restponse
        values of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A Call object for the RPC.
    """
    raise NotImplementedError()


class StreamUnaryMultiCallable(object):
  """Affords invoking a stream-unary RPC in any call style."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __call__(self, request_iterator, timeout):
    """Synchronously invokes the underlying RPC.

    Args:
      request_iterator: An iterator that yields request values for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      The response value for the RPC.

    Raises:
      exceptions.RpcError: Indicating that the RPC was aborted.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def future(self, request_iterator, timeout):
    """Asynchronously invokes the underlying RPC.

    Args:
      request_iterator: An iterator that yields request values for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A future.Future representing the RPC. In the event of RPC completion, the
        returned Future's result value will be the response value of the RPC.
        In the event of RPC abortion, the returned Future's exception value
        will be an exceptions.RpcError.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event(self, response_callback, abortion_callback, timeout):
    """Asynchronously invokes the underlying RPC.

    Args:
      request: The request value for the RPC.
      response_callback: A callback to be called to accept the restponse value
        of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A pair of a Call object for the RPC and a stream.Consumer to which the
        request values of the RPC should be passed.
    """
    raise NotImplementedError()


class StreamStreamMultiCallable(object):
  """Affords invoking a stream-stream RPC in any call style."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def __call__(self, request_iterator, timeout):
    """Synchronously invokes the underlying RPC.

    Args:
      request_iterator: An iterator that yields request values for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A CancellableIterator that yields the response values of the RPC and
        affords RPC cancellation. Drawing response values from the returned
        CancellableIterator may raise exceptions.RpcError indicating abortion
        of the RPC.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event(self, response_consumer, abortion_callback, timeout):
    """Asynchronously invokes the underlying RPC.

l    Args:
      response_consumer: A stream.Consumer to be called to accept the restponse
        values of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A pair of a Call object for the RPC and a stream.Consumer to which the
        request values of the RPC should be passed.
    """
    raise NotImplementedError()


class MethodImplementation(object):
  """A sum type that describes an RPC method implementation.

  Attributes:
    cardinality: A cardinality.Cardinality value.
    style: A style.Service value.
    unary_unary_inline: The implementation of the RPC method as a callable
      value that takes a request value and an RpcContext object and returns a
      response value. Only non-None if cardinality is
      cardinality.Cardinality.UNARY_UNARY and style is style.Service.INLINE.
    unary_stream_inline: The implementation of the RPC method as a callable
      value that takes a request value and an RpcContext object and returns an
      iterator of response values. Only non-None if cardinality is
      cardinality.Cardinality.UNARY_STREAM and style is style.Service.INLINE.
    stream_unary_inline: The implementation of the RPC method as a callable
      value that takes an iterator of request values and an RpcContext object
      and returns a response value. Only non-None if cardinality is
      cardinality.Cardinality.STREAM_UNARY and style is style.Service.INLINE.
    stream_stream_inline: The implementation of the RPC method as a callable
      value that takes an iterator of request values and an RpcContext object
      and returns an iterator of response values. Only non-None if cardinality
      is cardinality.Cardinality.STREAM_STREAM and style is
      style.Service.INLINE.
    unary_unary_event: The implementation of the RPC method as a callable value
      that takes a request value, a response callback to which to pass the
      response value of the RPC, and an RpcContext. Only non-None if
      cardinality is cardinality.Cardinality.UNARY_UNARY and style is
      style.Service.EVENT.
    unary_stream_event: The implementation of the RPC method as a callable
      value that takes a request value, a stream.Consumer to which to pass the
      the response values of the RPC, and an RpcContext. Only non-None if
      cardinality is cardinality.Cardinality.UNARY_STREAM and style is
      style.Service.EVENT.
    stream_unary_event: The implementation of the RPC method as a callable
      value that takes a response callback to which to pass the response value
      of the RPC and an RpcContext and returns a stream.Consumer to which the
      request values of the RPC should be passed. Only non-None if cardinality
      is cardinality.Cardinality.STREAM_UNARY and style is style.Service.EVENT.
    stream_stream_event: The implementation of the RPC method as a callable
      value that takes a stream.Consumer to which to pass the response values
      of the RPC and an RpcContext and returns a stream.Consumer to which the
      request values of the RPC should be passed. Only non-None if cardinality
      is cardinality.Cardinality.STREAM_STREAM and style is
      style.Service.EVENT.
  """
  __metaclass__ = abc.ABCMeta


class MultiMethodImplementation(object):
  """A general type able to service many RPC methods."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def service(self, name, response_consumer, context):
    """Services an RPC.

    Args:
      name: The RPC method name.
      response_consumer: A stream.Consumer to be called to accept the response
        values of the RPC.
      context: An RpcContext object.

    Returns:
      A stream.Consumer with which to accept the request values of the RPC. The
        consumer returned from this method may or may not be invoked to
        completion: in the case of RPC abortion, RPC Framework will simply stop
        passing values to this object. Implementations must not assume that this
        object will be called to completion of the request stream or even called
        at all.

    Raises:
      abandonment.Abandoned: May or may not be raised when the RPC has been
        aborted.
      exceptions.NoSuchMethodError: If this MultiMethod does not recognize the
        given RPC method name and is not able to service the RPC.
    """
    raise NotImplementedError()


class GenericStub(object):
  """Affords RPC methods to callers."""
  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def blocking_value_in_value_out(self, name, request, timeout):
    """Invokes a unary-request-unary-response RPC method.

    This method blocks until either returning the response value of the RPC
    (in the event of RPC completion) or raising an exception (in the event of
    RPC abortion).

    Args:
      name: The RPC method name.
      request: The request value for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      The response value for the RPC.

    Raises:
      exceptions.RpcError: Indicating that the RPC was aborted.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def future_value_in_value_out(self, name, request, timeout):
    """Invokes a unary-request-unary-response RPC method.

    Args:
      name: The RPC method name.
      request: The request value for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A future.Future representing the RPC. In the event of RPC completion, the
        returned Future will return an outcome indicating that the RPC returned
        the response value of the RPC. In the event of RPC abortion, the
        returned Future will return an outcome indicating that the RPC raised
        an exceptions.RpcError.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def inline_value_in_stream_out(self, name, request, timeout):
    """Invokes a unary-request-stream-response RPC method.

    Args:
      name: The RPC method name.
      request: The request value for the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A CancellableIterator that yields the response values of the RPC and
        affords RPC cancellation. Drawing response values from the returned
        CancellableIterator may raise exceptions.RpcError indicating abortion of
        the RPC.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def blocking_stream_in_value_out(self, name, request_iterator, timeout):
    """Invokes a stream-request-unary-response RPC method.

    This method blocks until either returning the response value of the RPC
    (in the event of RPC completion) or raising an exception (in the event of
    RPC abortion).

    Args:
      name: The RPC method name.
      request_iterator: An iterator that yields the request values of the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      The response value for the RPC.

    Raises:
      exceptions.RpcError: Indicating that the RPC was aborted.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def future_stream_in_value_out(self, name, request_iterator, timeout):
    """Invokes a stream-request-unary-response RPC method.

    Args:
      name: The RPC method name.
      request_iterator: An iterator that yields the request values of the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A future.Future representing the RPC. In the event of RPC completion, the
        returned Future will return an outcome indicating that the RPC returned
        the response value of the RPC. In the event of RPC abortion, the
        returned Future will return an outcome indicating that the RPC raised
        an exceptions.RpcError.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def inline_stream_in_stream_out(self, name, request_iterator, timeout):
    """Invokes a stream-request-stream-response RPC method.

    Args:
      name: The RPC method name.
      request_iterator: An iterator that yields the request values of the RPC.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A CancellableIterator that yields the response values of the RPC and
        affords RPC cancellation. Drawing response values from the returned
        CancellableIterator may raise exceptions.RpcError indicating abortion of
        the RPC.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event_value_in_value_out(
      self, name, request, response_callback, abortion_callback, timeout):
    """Event-driven invocation of a unary-request-unary-response RPC method.

    Args:
      name: The RPC method name.
      request: The request value for the RPC.
      response_callback: A callback to be called to accept the response value
        of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A Call object for the RPC.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event_value_in_stream_out(
      self, name, request, response_consumer, abortion_callback, timeout):
    """Event-driven invocation of a unary-request-stream-response RPC method.

    Args:
      name: The RPC method name.
      request: The request value for the RPC.
      response_consumer: A stream.Consumer to be called to accept the response
        values of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A Call object for the RPC.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event_stream_in_value_out(
      self, name, response_callback, abortion_callback, timeout):
    """Event-driven invocation of a unary-request-unary-response RPC method.

    Args:
      name: The RPC method name.
      response_callback: A callback to be called to accept the response value
        of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A pair of a Call object for the RPC and a stream.Consumer to which the
        request values of the RPC should be passed.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def event_stream_in_stream_out(
      self, name, response_consumer, abortion_callback, timeout):
    """Event-driven invocation of a unary-request-stream-response RPC method.

    Args:
      name: The RPC method name.
      response_consumer: A stream.Consumer to be called to accept the response
        values of the RPC.
      abortion_callback: A callback to be called and passed an Abortion value
        in the event of RPC abortion.
      timeout: A duration of time in seconds to allow for the RPC.

    Returns:
      A pair of a Call object for the RPC and a stream.Consumer to which the
        request values of the RPC should be passed.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def unary_unary_multi_callable(self, name):
    """Creates a UnaryUnaryMultiCallable for a unary-unary RPC method.

    Args:
      name: The RPC method name.

    Returns:
      A UnaryUnaryMultiCallable value for the named unary-unary RPC method.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def unary_stream_multi_callable(self, name):
    """Creates a UnaryStreamMultiCallable for a unary-stream RPC method.

    Args:
      name: The RPC method name.

    Returns:
      A UnaryStreamMultiCallable value for the name unary-stream RPC method.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def stream_unary_multi_callable(self, name):
    """Creates a StreamUnaryMultiCallable for a stream-unary RPC method.

    Args:
      name: The RPC method name.

    Returns:
      A StreamUnaryMultiCallable value for the named stream-unary RPC method.
    """
    raise NotImplementedError()

  @abc.abstractmethod
  def stream_stream_multi_callable(self, name):
    """Creates a StreamStreamMultiCallable for a stream-stream RPC method.

    Args:
      name: The RPC method name.

    Returns:
      A StreamStreamMultiCallable value for the named stream-stream RPC method.
    """
    raise NotImplementedError()


class DynamicStub(object):
  """A stub with RPC-method-bound multi-callable attributes.

  Instances of this type responsd to attribute access as follows: if the
  requested attribute is the name of a unary-unary RPC method, the value of the
  attribute will be a UnaryUnaryMultiCallable with which to invoke the RPC
  method; if the requested attribute is the name of a unary-stream RPC method,
  the value of the attribute will be a UnaryStreamMultiCallable with which to
  invoke the RPC method; if the requested attribute is the name of a
  stream-unary RPC method, the value of the attribute will be a
  StreamUnaryMultiCallable with which to invoke the RPC method; and if the
  requested attribute is the name of a stream-stream RPC method, the value of
  the attribute will be a StreamStreamMultiCallable with which to invoke the
  RPC method.
  """
  __metaclass__ = abc.ABCMeta



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Code for instructing systems under test to block or fail."""

import abc
import contextlib
import threading


class Control(object):
  """An object that accepts program control from a system under test.

  Systems under test passed a Control should call its control() method
  frequently during execution. The control() method may block, raise an
  exception, or do nothing, all according to the enclosing test's desire for
  the system under test to simulate hanging, failing, or functioning.
  """

  __metaclass__ = abc.ABCMeta

  @abc.abstractmethod
  def control(self):
    """Potentially does anything."""
    raise NotImplementedError()


class PauseFailControl(Control):
  """A Control that can be used to pause or fail code under control."""

  def __init__(self):
    self._condition = threading.Condition()
    self._paused = False
    self._fail = False

  def control(self):
    with self._condition:
      if self._fail:
        raise ValueError()

      while self._paused:
        self._condition.wait()

  @contextlib.contextmanager
  def pause(self):
    """Pauses code under control while controlling code is in context."""
    with self._condition:
      self._paused = True
    yield
    with self._condition:
      self._paused = False
      self._condition.notify_all()

  @contextlib.contextmanager
  def fail(self):
    """Fails code under control while controlling code is in context."""
    with self._condition:
      self._fail = True
    yield
    with self._condition:
      self._fail = False



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""A test to verify an implementation of the Face layer of RPC Framework."""

import abc
import unittest

from grpc.framework.face import interfaces
from grpc.framework.face.testing import callback as testing_callback
from grpc.framework.face.testing import control
from grpc.framework.face.testing import coverage
from grpc.framework.face.testing import digest
from grpc.framework.face.testing import stock_service
from grpc.framework.face.testing import test_case

_TIMEOUT = 3


class EventInvocationSynchronousEventServiceTestCase(
    test_case.FaceTestCase, coverage.FullCoverage):
  """A test of the Face layer of RPC Framework.

  Concrete subclasses must also extend unittest.TestCase.
  """
  __metaclass__ = abc.ABCMeta

  def setUp(self):
    """See unittest.TestCase.setUp for full specification.

    Overriding implementations must call this implementation.
    """
    self.control = control.PauseFailControl()
    self.digest = digest.digest(
        stock_service.STOCK_TEST_SERVICE, self.control, None)

    self.stub, self.memo = self.set_up_implementation(
        self.digest.name, self.digest.methods,
        self.digest.event_method_implementations, None)

  def tearDown(self):
    """See unittest.TestCase.tearDown for full specification.

    Overriding implementations must call this implementation.
    """
    self.tear_down_implementation(self.memo)

  def testSuccessfulUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        self.stub.event_value_in_value_out(
            name, request, callback.complete, callback.abort, _TIMEOUT)
        callback.block_until_terminated()
        response = callback.response()

        test_messages.verify(request, response, self)

  def testSuccessfulUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        self.stub.event_value_in_stream_out(
            name, request, callback, callback.abort, _TIMEOUT)
        callback.block_until_terminated()
        responses = callback.responses()

        test_messages.verify(request, responses, self)

  def testSuccessfulStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        callback = testing_callback.Callback()

        unused_call, request_consumer = self.stub.event_stream_in_value_out(
            name, callback.complete, callback.abort, _TIMEOUT)
        for request in requests:
          request_consumer.consume(request)
        request_consumer.terminate()
        callback.block_until_terminated()
        response = callback.response()

        test_messages.verify(requests, response, self)

  def testSuccessfulStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        callback = testing_callback.Callback()

        unused_call, request_consumer = self.stub.event_stream_in_stream_out(
            name, callback, callback.abort, _TIMEOUT)
        for request in requests:
          request_consumer.consume(request)
        request_consumer.terminate()
        callback.block_until_terminated()
        responses = callback.responses()

        test_messages.verify(requests, responses, self)

  def testSequentialInvocations(self):
    # pylint: disable=cell-var-from-loop
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        first_request = test_messages.request()
        second_request = test_messages.request()
        first_callback = testing_callback.Callback()
        second_callback = testing_callback.Callback()

        def make_second_invocation(first_response):
          first_callback.complete(first_response)
          self.stub.event_value_in_value_out(
              name, second_request, second_callback.complete,
              second_callback.abort, _TIMEOUT)

        self.stub.event_value_in_value_out(
            name, first_request, make_second_invocation, first_callback.abort,
            _TIMEOUT)
        second_callback.block_until_terminated()

        first_response = first_callback.response()
        second_response = second_callback.response()
        test_messages.verify(first_request, first_response, self)
        test_messages.verify(second_request, second_response, self)

  def testExpiredUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        with self.control.pause():
          self.stub.event_value_in_value_out(
              name, request, callback.complete, callback.abort, _TIMEOUT)
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.EXPIRED, callback.abortion())

  def testExpiredUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        with self.control.pause():
          self.stub.event_value_in_stream_out(
              name, request, callback, callback.abort, _TIMEOUT)
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.EXPIRED, callback.abortion())

  def testExpiredStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for unused_test_messages in test_messages_sequence:
        callback = testing_callback.Callback()

        self.stub.event_stream_in_value_out(
            name, callback.complete, callback.abort, _TIMEOUT)
        callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.EXPIRED, callback.abortion())

  def testExpiredStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        callback = testing_callback.Callback()

        unused_call, request_consumer = self.stub.event_stream_in_stream_out(
            name, callback, callback.abort, _TIMEOUT)
        for request in requests:
          request_consumer.consume(request)
        callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.EXPIRED, callback.abortion())

  def testFailedUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        with self.control.fail():
          self.stub.event_value_in_value_out(
              name, request, callback.complete, callback.abort, _TIMEOUT)
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.SERVICER_FAILURE, callback.abortion())

  def testFailedUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        with self.control.fail():
          self.stub.event_value_in_stream_out(
              name, request, callback, callback.abort, _TIMEOUT)
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.SERVICER_FAILURE, callback.abortion())

  def testFailedStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        callback = testing_callback.Callback()

        with self.control.fail():
          unused_call, request_consumer = self.stub.event_stream_in_value_out(
              name, callback.complete, callback.abort, _TIMEOUT)
          for request in requests:
            request_consumer.consume(request)
          request_consumer.terminate()
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.SERVICER_FAILURE, callback.abortion())

  def testFailedStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        callback = testing_callback.Callback()

        with self.control.fail():
          unused_call, request_consumer = self.stub.event_stream_in_stream_out(
              name, callback, callback.abort, _TIMEOUT)
          for request in requests:
            request_consumer.consume(request)
          request_consumer.terminate()
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.SERVICER_FAILURE, callback.abortion())

  def testParallelInvocations(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        first_request = test_messages.request()
        first_callback = testing_callback.Callback()
        second_request = test_messages.request()
        second_callback = testing_callback.Callback()

        self.stub.event_value_in_value_out(
            name, first_request, first_callback.complete, first_callback.abort,
            _TIMEOUT)
        self.stub.event_value_in_value_out(
            name, second_request, second_callback.complete,
            second_callback.abort, _TIMEOUT)
        first_callback.block_until_terminated()
        second_callback.block_until_terminated()

        first_response = first_callback.response()
        second_response = second_callback.response()
        test_messages.verify(first_request, first_response, self)
        test_messages.verify(second_request, second_response, self)

  @unittest.skip('TODO(nathaniel): implement.')
  def testWaitingForSomeButNotAllParallelInvocations(self):
    raise NotImplementedError()

  def testCancelledUnaryRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        with self.control.pause():
          call = self.stub.event_value_in_value_out(
              name, request, callback.complete, callback.abort, _TIMEOUT)
          call.cancel()
          callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.CANCELLED, callback.abortion())

  def testCancelledUnaryRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.unary_stream_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        request = test_messages.request()
        callback = testing_callback.Callback()

        call = self.stub.event_value_in_stream_out(
            name, request, callback, callback.abort, _TIMEOUT)
        call.cancel()
        callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.CANCELLED, callback.abortion())

  def testCancelledStreamRequestUnaryResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_unary_messages_sequences.iteritems()):
      for test_messages in test_messages_sequence:
        requests = test_messages.requests()
        callback = testing_callback.Callback()

        call, request_consumer = self.stub.event_stream_in_value_out(
            name, callback.complete, callback.abort, _TIMEOUT)
        for request in requests:
          request_consumer.consume(request)
        call.cancel()
        callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.CANCELLED, callback.abortion())

  def testCancelledStreamRequestStreamResponse(self):
    for name, test_messages_sequence in (
        self.digest.stream_stream_messages_sequences.iteritems()):
      for unused_test_messages in test_messages_sequence:
        callback = testing_callback.Callback()

        call, unused_request_consumer = self.stub.event_stream_in_stream_out(
            name, callback, callback.abort, _TIMEOUT)
        call.cancel()
        callback.block_until_terminated()

        self.assertEqual(interfaces.Abortion.CANCELLED, callback.abortion())



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""Utilities helpful for working with the base layer of RPC Framework."""

import collections
import threading

from grpc.framework.base import interfaces


class _ServicedSubscription(
    collections.namedtuple('_ServicedSubscription', ['kind', 'ingestor']),
    interfaces.ServicedSubscription):
  """See interfaces.ServicedSubscription for specification."""

_NONE_SUBSCRIPTION = _ServicedSubscription(
    interfaces.ServicedSubscription.Kind.NONE, None)
_TERMINATION_ONLY_SUBSCRIPTION = _ServicedSubscription(
    interfaces.ServicedSubscription.Kind.TERMINATION_ONLY, None)


def none_serviced_subscription():
  """Creates a "none" interfaces.ServicedSubscription object.

  Returns:
    An interfaces.ServicedSubscription indicating no subscription to an
      operation's results (such as would be the case for a fire-and-forget
      operation invocation).
  """
  return _NONE_SUBSCRIPTION


def termination_only_serviced_subscription():
  """Creates a "termination only" interfaces.ServicedSubscription object.

  Returns:
    An interfaces.ServicedSubscription indicating that the front-side customer
      is interested only in the overall termination outcome of the operation
      (such as completion or expiration) and would ignore the actual results of
      the operation.
  """
  return _TERMINATION_ONLY_SUBSCRIPTION


def full_serviced_subscription(ingestor):
  """Creates a "full" interfaces.ServicedSubscription object.

  Args:
    ingestor: An interfaces.ServicedIngestor.

  Returns:
    An interfaces.ServicedSubscription object indicating a full
      subscription.
  """
  return _ServicedSubscription(
      interfaces.ServicedSubscription.Kind.FULL, ingestor)


def wait_for_idle(end):
  """Waits for an interfaces.End to complete all operations.

  Args:
    end: Any interfaces.End.
  """
  event = threading.Event()
  end.add_idle_action(event.set)
  event.wait()



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""State and behavior for translating between sync and async control flow."""

import threading

from grpc.framework.base import interfaces as base_interfaces
from grpc.framework.face import exceptions
from grpc.framework.face import interfaces
from grpc.framework.foundation import abandonment
from grpc.framework.foundation import stream

INTERNAL_ERROR_LOG_MESSAGE = ':-( RPC Framework (Face) Internal Error! :-('

_OPERATION_OUTCOME_TO_RPC_ABORTION = {
    base_interfaces.Outcome.CANCELLED: interfaces.Abortion.CANCELLED,
    base_interfaces.Outcome.EXPIRED: interfaces.Abortion.EXPIRED,
    base_interfaces.Outcome.RECEPTION_FAILURE:
        interfaces.Abortion.NETWORK_FAILURE,
    base_interfaces.Outcome.TRANSMISSION_FAILURE:
        interfaces.Abortion.NETWORK_FAILURE,
    base_interfaces.Outcome.SERVICED_FAILURE:
        interfaces.Abortion.SERVICED_FAILURE,
    base_interfaces.Outcome.SERVICER_FAILURE:
        interfaces.Abortion.SERVICER_FAILURE,
}


def _as_operation_termination_callback(rpc_abortion_callback):
  def operation_termination_callback(operation_outcome):
    rpc_abortion = _OPERATION_OUTCOME_TO_RPC_ABORTION.get(
        operation_outcome, None)
    if rpc_abortion is not None:
      rpc_abortion_callback(rpc_abortion)
  return operation_termination_callback


def _abortion_outcome_to_exception(abortion_outcome):
  if abortion_outcome == base_interfaces.Outcome.CANCELLED:
    return exceptions.CancellationError()
  elif abortion_outcome == base_interfaces.Outcome.EXPIRED:
    return exceptions.ExpirationError()
  elif abortion_outcome == base_interfaces.Outcome.SERVICER_FAILURE:
    return exceptions.ServicerError()
  elif abortion_outcome == base_interfaces.Outcome.SERVICED_FAILURE:
    return exceptions.ServicedError()
  else:
    return exceptions.NetworkError()


class UnaryConsumer(stream.Consumer):
  """A stream.Consumer that should only ever be passed one value."""

  def __init__(self, on_termination):
    self._on_termination = on_termination
    self._value = None

  def consume(self, value):
    self._value = value

  def terminate(self):
    self._on_termination(self._value)

  def consume_and_terminate(self, value):
    self._on_termination(value)


class Rendezvous(stream.Consumer):
  """A rendez-vous with stream.Consumer and iterator interfaces."""

  def __init__(self):
    self._condition = threading.Condition()
    self._values = []
    self._values_completed = False
    self._abortion = None

  def consume(self, value):
    with self._condition:
      self._values.append(value)
      self._condition.notify()

  def terminate(self):
    with self._condition:
      self._values_completed = True
      self._condition.notify()

  def consume_and_terminate(self, value):
    with self._condition:
      self._values.append(value)
      self._values_completed = True
      self._condition.notify()

  def __iter__(self):
    return self

  def next(self):
    with self._condition:
      while ((self._abortion is None) and
             (not self._values) and
             (not self._values_completed)):
        self._condition.wait()
      if self._abortion is not None:
        raise _abortion_outcome_to_exception(self._abortion)
      elif self._values:
        return self._values.pop(0)
      elif self._values_completed:
        raise StopIteration()
      else:
        raise AssertionError('Unreachable code reached!')

  def set_outcome(self, outcome):
    with self._condition:
      if outcome is not base_interfaces.Outcome.COMPLETED:
        self._abortion = outcome
        self._condition.notify()


class RpcContext(interfaces.RpcContext):
  """A wrapped base_interfaces.OperationContext."""

  def __init__(self, operation_context):
    self._operation_context = operation_context

  def is_active(self):
    return self._operation_context.is_active()

  def time_remaining(self):
    return self._operation_context.time_remaining()

  def add_abortion_callback(self, abortion_callback):
    self._operation_context.add_termination_callback(
        _as_operation_termination_callback(abortion_callback))


def pipe_iterator_to_consumer(iterator, consumer, active, terminate):
  """Pipes values emitted from an iterator to a stream.Consumer.

  Args:
    iterator: An iterator from which values will be emitted.
    consumer: A stream.Consumer to which values will be passed.
    active: A no-argument callable that returns True if the work being done by
      this function is still valid and should not be abandoned and False if the
      work being done by this function should be abandoned.
    terminate: A boolean indicating whether or not this function should
      terminate the given consumer after passing to it all values emitted by the
      given iterator.

  Raises:
    abandonment.Abandoned: If this function quits early after seeing False
      returned by the active function passed to it.
    Exception: This function raises whatever exceptions are raised by iterating
      over the given iterator.
  """
  for element in iterator:
    if not active():
      raise abandonment.Abandoned()

    consumer.consume(element)

    if not active():
      raise abandonment.Abandoned()
  if terminate:
    consumer.terminate()


def abortion_outcome_to_exception(abortion_outcome):
  return _abortion_outcome_to_exception(abortion_outcome)


def as_operation_termination_callback(rpc_abortion_callback):
  return _as_operation_termination_callback(rpc_abortion_callback)



# Copyright 2015, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# TODO(nathaniel): Remove this from source control after having made
# generation from the stock.proto source part of GRPC's build-and-test
# process.

# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: stock.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='stock.proto',
  package='stock',
  serialized_pb=_b('\n\x0bstock.proto\x12\x05stock\">\n\x0cStockRequest\x12\x0e\n\x06symbol\x18\x01 \x01(\t\x12\x1e\n\x13num_trades_to_watch\x18\x02 \x01(\x05:\x01\x30\"+\n\nStockReply\x12\r\n\x05price\x18\x01 \x01(\x02\x12\x0e\n\x06symbol\x18\x02 \x01(\t2\x96\x02\n\x05Stock\x12=\n\x11GetLastTradePrice\x12\x13.stock.StockRequest\x1a\x11.stock.StockReply\"\x00\x12I\n\x19GetLastTradePriceMultiple\x12\x13.stock.StockRequest\x1a\x11.stock.StockReply\"\x00(\x01\x30\x01\x12?\n\x11WatchFutureTrades\x12\x13.stock.StockRequest\x1a\x11.stock.StockReply\"\x00\x30\x01\x12\x42\n\x14GetHighestTradePrice\x12\x13.stock.StockRequest\x1a\x11.stock.StockReply\"\x00(\x01')
)
_sym_db.RegisterFileDescriptor(DESCRIPTOR)




_STOCKREQUEST = _descriptor.Descriptor(
  name='StockRequest',
  full_name='stock.StockRequest',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='symbol', full_name='stock.StockRequest.symbol', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='num_trades_to_watch', full_name='stock.StockRequest.num_trades_to_watch', index=1,
      number=2, type=5, cpp_type=1, label=1,
      has_default_value=True, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=22,
  serialized_end=84,
)


_STOCKREPLY = _descriptor.Descriptor(
  name='StockReply',
  full_name='stock.StockReply',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='price', full_name='stock.StockReply.price', index=0,
      number=1, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='symbol', full_name='stock.StockReply.symbol', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=86,
  serialized_end=129,
)

DESCRIPTOR.message_types_by_name['StockRequest'] = _STOCKREQUEST
DESCRIPTOR.message_types_by_name['StockReply'] = _STOCKREPLY

StockRequest = _reflection.GeneratedProtocolMessageType('StockRequest', (_message.Message,), dict(
  DESCRIPTOR = _STOCKREQUEST,
  __module__ = 'stock_pb2'
  # @@protoc_insertion_point(class_scope:stock.StockRequest)
  ))
_sym_db.RegisterMessage(StockRequest)

StockReply = _reflection.GeneratedProtocolMessageType('StockReply', (_message.Message,), dict(
  DESCRIPTOR = _STOCKREPLY,
  __module__ = 'stock_pb2'
  # @@protoc_insertion_point(class_scope:stock.StockReply)
  ))
_sym_db.RegisterMessage(StockReply)


# @@protoc_insertion_point(module_scope)
